#include <QAction>
#include <QApplication>
#include <QMainWindow>
#include <QMenuBar>
#include <QMenu>
#include <QPushButton>
#include <QStyle>
#include <QUrl>

#include <QDesktopServices>
#include <QScreen>

#include <QDebug>
#include <QTimer>

#include <algorithm>
#include <iostream>

static QString title()
{
    return QLatin1String(QT_VERSION_STR) + QLatin1Char(' ') + QGuiApplication::platformName()
        + QLatin1Char(' ') + QApplication::style()->objectName();
}

class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(QWidget *parent = nullptr);

private slots:
    void storeGeo();
    void restoreGeo();
    void setChildGeometry();

private:
    void createChildWindow();
    void destoryChildWindow();

    QWidget *m_childWindow;
    QByteArray m_storedGeometry;
};


MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent)
{
    setObjectName("Top");
    setWindowTitle(title());

    createChildWindow();  // same as: childWindow = new QWidget(this, Qt::Window);
    setChildGeometry(); // same as: childWindow->setGeometry(60, 60, 800, 800);
    m_childWindow->show();

    // Major regression since 5.9.4 - Window geometry needs to be called after show.
    // childWindow->setGeometry(60, 60, 800, 800);

    qDebug() << m_childWindow->geometry();
    // return; - returning now is sufficent to show one major regression.

    QMenu *menu = menuBar()->addMenu("Store/Restore");

    menu->addAction("Store child geometry", this, &MainWindow::storeGeo);
    menu->addAction("Restore child geometry", this, &MainWindow::restoreGeo);
    menu->addAction("Set child geometry", this, &MainWindow::setChildGeometry);
    menu->addAction("Set child geometry", this, &MainWindow::setChildGeometry);
    auto qa = menu->addAction("Quit", this, QCoreApplication::quit);
    qa->setShortcut(Qt::CTRL + Qt::Key_Q);

    // setGeometry(1200, 60, 500, 500);
}

void MainWindow::storeGeo()
{
    if (!m_childWindow)
        return;
    qDebug() << __FUNCTION__ << "Store geometry size" << m_childWindow->geometry();
    m_storedGeometry = m_childWindow->saveGeometry();
    // If we just hide the widget then the restore bug does not happen.
    destoryChildWindow();
}

void MainWindow::restoreGeo()
{
    if (!m_childWindow)
        createChildWindow();

    // In 5.9.4 there seems to be issues if the window is shown before restore geometry was called
    // In 5.12.3 this restores a wrong geometry
    qDebug() << __FUNCTION__;
    m_childWindow->restoreGeometry(m_storedGeometry);
    m_childWindow->show();
    qDebug() << "Restore geometry size" << m_childWindow->geometry();
}

void MainWindow::createChildWindow()
{
    qDebug() << __FUNCTION__;
    // It however does not matter if the child has a parent.
    m_childWindow = new QWidget(this, Qt::Window);
    m_childWindow->setObjectName("Child");
    m_childWindow->setWindowTitle("Child - " + windowTitle());
    auto b = new QPushButton("bug", m_childWindow);
    connect(b, &QAbstractButton::clicked, b,
            [] ()
    {
        auto u = "https://bugreports.qt.io/browse/QTBUG-75940";
        qDebug() << __FUNCTION__ << u;
        QDesktopServices::openUrl(QUrl(u));
    });
}

void MainWindow::destoryChildWindow()
{
    qDebug() << __FUNCTION__;
    m_childWindow->deleteLater();
    m_childWindow = nullptr;
}

void MainWindow::setChildGeometry()
{
    qDebug() << __FUNCTION__;
    m_childWindow->setGeometry(60, 60, 800, 800);
    qDebug() << __FUNCTION__ << '<' << m_childWindow->geometry();
}

int main(int argc, char *argv[])
{
    QStringList args;
    std::copy(argv + 1, argv + argc, std::back_inserter(args));

    if (args.contains("-n"))  {
        QCoreApplication::setAttribute(Qt::AA_DisableHighDpiScaling);
    } else {
        QCoreApplication::setAttribute(Qt::AA_UseHighDpiPixmaps);
        QCoreApplication::setAttribute(Qt::AA_EnableHighDpiScaling);
    }

    QApplication a(argc, argv);

    QString t = title();
    {
        QDebug d = qDebug().noquote().nospace();
        d <<  t << ' ' << args << '\n';
        for (auto s : QGuiApplication::screens()) {
            d <<  s->name() << ' ' << s->geometry() << ' ' << s->logicalDotsPerInchX() << "dpi, dpr="
               << s->devicePixelRatio() << '\n';
        }
    }

    MainWindow w;
    w.show();

    return a.exec();
}

#include "main.moc"
